import numpy as np
import pytest

from brad import permutation_test
from brad.exceptions import EmptyDataError

from .helpers import safe_mean

RANDOM_LABELS = np.random.binomial(1, 0.5, size=(20,))
RANDOM_DATA = np.random.normal(RANDOM_LABELS)
LABELS = np.array([1, 2, 4, 125, 1261, 0.1])


def mean_diff(x, y):
    return safe_mean(x[y == 1]) - safe_mean(x[y == 0])


def make_check_permutation(labels):
    def check_permutation(_, y):
        return set(labels[~np.isnan(labels)].tolist()) == set(
            y[~np.isnan(y)].tolist()
        )

    return check_permutation


def test_shape():
    for n in [2, 7, 102]:
        result = permutation_test(
            RANDOM_DATA, RANDOM_LABELS, n, statistic=mean_diff
        )
        assert result.samples.shape == (n,)


def test_empty_dataset():
    for n in [2, 7, 102]:
        empty_dataset = np.arange(0)
        check_permutation = make_check_permutation(empty_dataset)
        with pytest.raises(
            EmptyDataError, match="Supplied data is empty! Nothing to sample"
        ):
            permutation_test(
                empty_dataset, empty_dataset, n, statistic=check_permutation
            )


def test_permutation_test_samples_lables_permutations():
    check_permutation = make_check_permutation(LABELS)
    result = permutation_test(
        np.arange(6), LABELS, 1000, statistic=check_permutation
    )
    assert result.samples.all()


def test_data_labels_mismatch():
    with pytest.raises(ValueError):
        permutation_test(np.arange(10), np.arange(11), 10)



import numpy as np
import pytest

from brad import smooth_bootstrap
from brad.exceptions import EmptyDataError

from .helpers import RANDOM_10x5, safe_mean

KERNELS = [
    "gaussian",
    "tophat",
]
KERNELS_WITHOUT_SAMPLE = ["epanechnikov", "exponential", "linear", "cosine"]


@pytest.mark.parametrize("n_samples", [2, 7, 102])
@pytest.mark.parametrize("kernel", KERNELS)
@pytest.mark.parametrize("bandwidth", [0.1, 0.2, 0.5])
def test_shape(n_samples, kernel, bandwidth):
    data = RANDOM_10x5
    shape = data.shape

    assert (
        smooth_bootstrap(
            data, n_samples, kernel=kernel, bandwidth=bandwidth
        ).samples.shape
        == (n_samples,) + shape
    )

    assert (
        smooth_bootstrap(
            data,
            n_samples,
            kernel=kernel,
            bandwidth=bandwidth,
            statistic=safe_mean,
        ).samples.shape
        == (n_samples,)
    )


@pytest.mark.parametrize("kernel", KERNELS_WITHOUT_SAMPLE)
def test_sklearn_kernal_no_sample(kernel):
    """
    Smooth bootstrap uses scikit-learn's KDE implementation, but requires that
    we can draw samples, and this is not possible for many of the available
    kernels. Raise an error if a choice is made that does not allow samples to
    be drawn.
    """
    data = RANDOM_10x5

    with pytest.raises(ValueError):
        smooth_bootstrap(data, 1, kernel=kernel)


def test_invalid_kernel():
    data = RANDOM_10x5

    with pytest.raises(ValueError):
        smooth_bootstrap(data, 1, kernel="invalid-kernel")


@pytest.mark.parametrize("n_samples", [2, 7, 102])
def test_empty_dataset(n_samples):
    empty_dataset = np.arange(0)

    with pytest.raises(
        EmptyDataError, match="Supplied data is empty! Nothing to sample"
    ):
        smooth_bootstrap(empty_dataset, n_samples)


@pytest.mark.parametrize("n_samples", [2, 7, 102])
def test_scalar_dataset(n_samples):
    scalar_dataset = np.array(0)

    with pytest.raises(
        ValueError, match="Parameter 'data' cannot be a scalar"
    ):
        smooth_bootstrap(scalar_dataset, n_samples)



import abc
from typing import Iterator

import numpy as np
from numpy.core.numeric import asanyarray
from scipy.stats.distributions import (
    beta,
    chi2,
    expon,
    f,
    gamma,
    lognorm,
    norm,
    t,
)
from sklearn.neighbors import KernelDensity

from brad.exceptions import EmptyDataError

__all__ = ["NonParametricSampler", "ParametricSampler", "SmoothSampler"]

# all scipy distributions have loc and scale parameters that allow you to
# shift and stretch the distributions. For now, we force loc and scale to
# be 0 and 1 resp. so that standard versions of the distributions are fit
_DIST_LOOKUP = {
    "beta": [beta, {"floc": 0, "fscale": 1}],
    "chi2": [chi2, {"floc": 0, "fscale": 1}],
    "expon": [expon, {"fscale": 1}],
    "f": [f, {"floc": 0, "fscale": 1}],
    "gamma": [gamma, {"floc": 0}],
    "lognorm": [lognorm, {"fscale": 1}],
    "norm": [norm, {}],
    "t": [t, {"floc": 0, "fscale": 1}],
}


class Sampler(abc.ABC):
    """
    Sampler base class. Samplers generate samples from a data set with their
    __iter__ method. They should be able to yield samples indefinitely, things
    like the sample_size are controlled at a higher level (hopefully to make
    parallelisation easier).
    """

    def __init__(self, data: np.ndarray) -> None:
        """
        Args:
            data: Dataset to draw samples from. Samples are taken along the
                zeroth dimension.
        """
        self._data = asanyarray(data)

        if self._data.ndim == 0:
            raise ValueError("Parameter 'data' cannot be a scalar")
        elif self._data.size == 0:
            raise EmptyDataError("Supplied data is empty! Nothing to sample")

        self.sample_size = self._data.shape[0]

    @abc.abstractclassmethod
    def __iter__(self) -> Iterator[np.ndarray]:
        """This method should yield samples from the sampler"""


class NonParametricSampler(Sampler):
    """
    Constructs samples by sampling the original dataset with replacement. Used
    for computing the non-parametric bootstrap.
    """

    def __init__(
        self, data: np.ndarray, *, max_buffer: int = 1_000_000
    ) -> None:
        """
        Args:
            data: Dataset to draw samples from. Samples are taken along the
                zeroth dimension.
            max_buffer: Indices for resampling are sampled in large batches and
                stored in a buffer for efficiency. This argument sets the size
                of that buffer.
        """
        super().__init__(data)
        self.max_buffer = max(max_buffer, self.sample_size)

    def __iter__(self) -> Iterator[np.ndarray]:
        buffer_size = self.max_buffer - (self.max_buffer % self.sample_size)

        rng = np.random.default_rng()

        while True:
            buffer = rng.choice(self.sample_size, buffer_size)

            for batch in np.split(buffer, buffer_size // self.sample_size):
                yield self._data[batch]


class ParametricSampler(Sampler):
    """
    Fits a parametric approximation to the data, then draws samples from the
    approximation. Used for computing the parametric bootstrap.
    """

    def __init__(self, data: np.ndarray, *, family: str) -> None:
        """
        Args:
            data: Dataset to fit parametric approximation to. Must be 1-dimensional.
            family: Parametric family to fit data to. Options are: `beta`,
                `chi2`, `expon`, `f`, `gamma`, `lognorm`, `norm`, `t`.
        """
        super().__init__(data)

        if self._data.ndim != 1:
            raise ValueError("Parametric samplers expect 1-d data.")

        if family in _DIST_LOOKUP:
            self._dist, self._fit_kwargs = _DIST_LOOKUP[family]
        else:
            raise ValueError(
                f"Invalid option '{family}' for parameter 'family'"
            )

        self._args = self._dist.fit(self._data, **self._fit_kwargs)

    def __iter__(self) -> Iterator[np.ndarray]:
        while True:
            yield self._dist.rvs(*self._args, size=self.sample_size)


class SmoothSampler(Sampler):
    """
    Fits a kernel density estimator to the data, then draws samples from the
    approximation. Used for computing the smooth bootstrap.
    """

    def __init__(
        self,
        data: np.ndarray,
        *,
        kernel: str = "gaussian",
        bandwidth: float = 0.2,
    ) -> None:
        """
        Args:
            data: Dataset to fit smooth approximation to.
            kernel: Kernel to use in the KDE. Options are `gaussian` or
                `tophat`.
            bandwidth: Bandwidth of the KDE.
        """
        if kernel not in {"gaussian", "tophat"}:
            raise ValueError(
                f"Invalid option '{kernel}' for parameter 'kernel'"
            )

        super().__init__(data)
        self.kde = KernelDensity(kernel=kernel, bandwidth=bandwidth).fit(
            self._data
        )

    def __iter__(self) -> Iterator[np.ndarray]:
        while True:
            yield self.kde.sample(n_samples=self.sample_size)


class JackknifeSampler(Sampler):
    """
    Generates samples by a leave-one-out strategy. A little different to the
    other samplers as will only yield finitely many samples.
    """

    def __iter__(self):
        for i in range(self.sample_size):
            yield np.delete(self._data, i, axis=0)


class BootknifeSampler(Sampler):
    """
    Generates samples by leaving out one datapoint, then sampling with
    replacement N times from the remaining N - 1 datapoints. Used for
    calculating bootknife estimates, which are similar to non-parametric
    bootstrap estimates, but with reduced bias.

    Bootknife sees best results with stratified samples (i.e. when each
    datapoint is left out equally often, or approximately so if the number of
    samples is not a multiple of the number of datapoints). This sampler does
    a stratification like this, but randomises the order of the datapoints so
    that there isn't a systematic bias toward the early datapoints should
    multiple samplers be created and their samples combined.
    """

    def __init__(self, data: np.ndarray, *, max_buffer=1_000_000) -> None:
        """
        Args:
            data: Dataset to draw samples from. Samples are taken along the
                zeroth dimension.
            max_buffer: Indices for resampling are sampled in large batches and
                stored in a buffer for efficiency. This argument sets the size
                of that buffer.
        """
        super().__init__(data)
        self.max_buffer = max_buffer

    def __iter__(self) -> Iterator[np.ndarray]:
        buffer_size = self.max_buffer - (self.max_buffer % self.sample_size)

        rng = np.random.default_rng()

        # to ensure good stratification we randomise the order of the features
        # then delete them in order. stratification will be imperfect when done
        # in parallel by multiple workers, but it will be pretty good.
        order = rng.permutation(self.sample_size)
        idx = 0

        while True:
            buffer = rng.choice(self.sample_size - 1, buffer_size)

            for batch in np.split(buffer, buffer_size // self.sample_size):
                yield np.delete(self._data, order[idx], axis=0)[batch]
                idx = (idx + 1) % self.sample_size


class PermutationSampler(Sampler):
    """
    Generates samples by a permuting the dataset. Used to conduct permutation
    tests by sampling permutations of the labels in a labelled dataset.
    """

    def __iter__(self) -> Iterator[np.ndarray]:
        rng = np.random.default_rng()

        while True:
            yield self._data[rng.permutation(self.sample_size)]



import numpy as np
import pytest

from brad import bootknife
from brad.exceptions import EmptyDataError

from .helpers import RANDOM_10x5, safe_mean

TUPLES = ((1, 2), (1, 6), (4, 7))
TUPLES_SET = set(TUPLES)


@pytest.mark.parametrize("n_samples", [2, 7, 102])
def test_shape(n_samples):
    data = RANDOM_10x5
    shape = data.shape

    assert bootknife(data, n_samples).samples.shape == (n_samples,) + shape
    assert bootknife(data, n_samples, statistic=safe_mean).samples.shape == (
        n_samples,
    )


def test_bootknife_empty_dataset():
    empty_dataset = np.arange(0)

    with pytest.raises(
        EmptyDataError, match="Supplied data is empty! Nothing to sample"
    ):
        bootknife(empty_dataset, 1)


def test_bootknife_scalar_dataset():
    scalar_dataset = np.array(0)

    with pytest.raises(
        ValueError, match="Parameter 'data' cannot be a scalar"
    ):
        bootknife(scalar_dataset, 1)


def test_bootknife_returned_values_are_samples():
    samples = bootknife(TUPLES, 10).samples
    assert all(
        [
            tuple(samples[i, j]) in TUPLES_SET
            for i in range(10)
            for j in range(3)
        ]
    )



import functools
from typing import Callable

import numpy as np
from numpy.core.numeric import asanyarray

from brad.models import Resample
from brad.samplers import (
    BootknifeSampler,
    JackknifeSampler,
    NonParametricSampler,
    ParametricSampler,
    PermutationSampler,
    Sampler,
    SmoothSampler,
)


def resample(
    sampler: Sampler,
    n_samples: int,
    *,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Draw `n_samples` from `sampler` and optionally apply `statistic` to each
    sample. Used as a building block for all other resampling functions.

    Args:
        n_samples: The number of samples to draw.
        sampler: A sampler from which samples will be drawn.
        statistic: A statistic that can be applied to each sample. It not
            provided, then samples are returned untouched.

    Returns:
        Data and resamples from the sampler.
    """
    statistic = statistic if statistic is not None else (lambda sample: sample)
    data = asanyarray(statistic(sampler._data))
    samples = np.zeros((n_samples,) + data.shape)

    for i, sample in enumerate(sampler):
        if i >= n_samples:
            break
        samples[i] = asanyarray(statistic(sample))

    return Resample(data, samples)


def bootstrap(
    data: np.ndarray,
    n_samples: int,
    *,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Generate bootstrap samples from data.

    Args:
        data: Input data from which to draw bootstrap samples.
    n_samples: The number of samples to draw.
    statistic: A function/statistic to apply to each bootstrap sample. If
        statistic is supplied then the returned array will have shape
        `(n_samples, ) + statistic_shape`.

    Returns:
        Data and non-parametric bootstrap resamples.
    """
    sampler = NonParametricSampler(data)
    return resample(sampler, n_samples, statistic=statistic)


def smooth_bootstrap(
    data: np.ndarray,
    n_samples: int,
    *,
    kernel: str = "gaussian",
    bandwidth: float = 0.2,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Generate smooth bootstrap samples from data. First fits a KDE to the
    provided data, then samples from the fitted smooth approximation.

    Args:
        data: Input data to which a KDE will be fit.
        n_samples: The number of sampels to draw.
        kernel: Kernel to use when fitting the KDE. Can choose
            either 'gaussian' or 'tophat'.
        bandwidth: Bandwidth of the kernel used in the KDE.
        statistic: A function/statistic to apply to each bootstrap
            sample. If statistic is supplied then the returned array will have
            shape `(n_samples, ) + statistic_shape`.

    Returns:
        Data and smooth bootstrap resamples.
    """
    sampler = SmoothSampler(data, kernel=kernel, bandwidth=bandwidth)
    return resample(sampler, n_samples, statistic=statistic)


def parametric_bootstrap(
    data: np.ndarray,
    n_samples: int,
    *,
    family: str,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Generate parametric bootstrap samples from data. First fits a parametric
    distribution to the provided data, then samples from the fitted parametric
    approximation.

    Args:
        data: Input data to which a KDE will be fit.
        n_samples: The number of sampels to draw.
        family: The parametric family from which to select the
            approximation. Options are: beta, chi2, expon, f, gamma, lognorm,
            norm, t.
        statistic: A function/statistic to apply to each bootstrap
            sample. If statistic is supplied then the returned array will have
            shape `(n_samples, ) + statistic_shape`.

    Returns:
        Data and smooth bootstrap resamples.
    """
    sampler = ParametricSampler(data, family=family)
    return resample(sampler, n_samples, statistic=statistic)


def permutation_test(
    data: np.ndarray,
    labels: np.ndarray,
    n_samples: int,
    *,
    statistic: Callable[[np.ndarray, np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Randomly permute labels and apply statistic n_samples times.

    Args:
        data: Input data with which to conduct permutation test.
        labels: Labels must have x.shape[0] == labels.shape[0].
        n_samples: The number of samples / permutations to draw.
        statistic: A function to apply to each bootstrap sample. The expected
            signature is `(x, labels) -> stat`.

    Returns:
        An array of samples with shape `(n_samples, ) + statistic_shape`.
    """
    if data.shape[0] != labels.shape[0]:
        raise ValueError(
            "Shape mismatch: first dimension of data and labels should be the "
            "same size."
        )

    sampler = PermutationSampler(labels)

    statistic = functools.partial(
        statistic or (lambda data, labels: labels), data
    )
    return resample(sampler, n_samples, statistic=statistic)


def bootknife(
    data: np.ndarray,
    n_samples: int,
    *,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Generate bootknife samples from data.

    Args:
        data: Input data from which to draw bootstrap samples.
        n_samples: The number of samples to draw.
        statistic:
            A function/statistic to apply to each bootstrap sample. If
            statistic is supplied then the returned array will have shape
            `(n_samples, ) + statistic_shape`.

    Returns:
        An array of samples.
    """
    sampler = BootknifeSampler(data)
    return resample(sampler, n_samples, statistic=statistic)


def jackknife(
    data: np.ndarray,
    *,
    statistic: Callable[[np.ndarray], np.ndarray] = None,
) -> Resample:
    """
    Generate jackknife samples from data.

    Args:
        data: Input data from which to draw jackknife samples.
        statistic: A function/statistic to apply to each sample. If
            statistic is supplied then the returned array will have shape
            `(n_samples, ) + statistic_shape`.

    Returns:
        An array of samples.
    """
    sampler = JackknifeSampler(data)
    # TODO: this is broken
    statistic = (
        statistic
        if statistic is not None
        else (lambda sample: sample[: sampler._data.shape[0] - 1])
    )
    return resample(sampler, len(data), statistic=statistic)



from brad.resample import (
    bootknife,
    bootstrap,
    jackknife,
    parametric_bootstrap,
    permutation_test,
    smooth_bootstrap,
)

try:
    from importlib.metadata import (  # type: ignore
        PackageNotFoundError,
        version,
    )
except ModuleNotFoundError:
    # if using Python 3.7, import from the backport
    from importlib_metadata import (  # type: ignore
        PackageNotFoundError,
        version,
    )

try:
    __version__ = version("brad")
except PackageNotFoundError:
    # package is not installed
    pass


__all__ = [
    "bootknife",
    "bootstrap",
    "jackknife",
    "parametric_bootstrap",
    "permutation_test",
    "smooth_bootstrap",
]



import numpy as np
import pytest

from brad import jackknife
from brad.exceptions import EmptyDataError

from .helpers import RANDOM_10x5, safe_mean

TUPLES = ((1, 2), (1, 6), (4, 7))
TUPLES_SET = set(TUPLES)


def test_shape():
    shape = RANDOM_10x5.shape

    assert (
        jackknife(RANDOM_10x5).samples.shape
        == (shape[0], shape[0] - 1) + shape[1:]
    )
    assert jackknife(RANDOM_10x5, statistic=safe_mean).samples.shape == (
        shape[0],
    )


def test_bootknife_empty_dataset():
    empty_dataset = np.arange(0)
    with pytest.raises(
        EmptyDataError, match="Supplied data is empty! Nothing to sample"
    ):
        jackknife(empty_dataset)


def test_bootknife_returned_values_are_samples():
    samples = jackknife(TUPLES).samples
    assert all(
        [
            tuple(samples[i, j]) in TUPLES_SET
            for i in range(3)
            for j in range(2)
        ]
    )



import numpy as np
import pytest
from scipy.stats.distributions import (
    beta,
    chi2,
    expon,
    f,
    gamma,
    lognorm,
    norm,
    t,
)

from brad import parametric_bootstrap
from brad.exceptions import EmptyDataError

from .helpers import RANDOM_10x1, RANDOM_10x5, safe_mean

dist_names = [d.name for d in [lognorm, norm]]

DISTRIBUTIONS = [beta, chi2, expon, f, gamma, lognorm, norm, t]
DISTRIBUTION_PARAMETERS = [
    (3, 5),
    (5,),
    (5,),
    (50, 50),
    (6, 0, 4),
    (5, 1),
    (3, 2),
    (4,),
]
FAMILY_DATA = [
    (d.name, d.rvs(*params, size=1000))
    for d, params in zip(DISTRIBUTIONS, DISTRIBUTION_PARAMETERS)
]


# F distribution spits out a runtime warning when fitting
@pytest.mark.filterwarnings("ignore::RuntimeWarning")
@pytest.mark.parametrize("n_samples", [2, 7, 102])
@pytest.mark.parametrize("family,data", FAMILY_DATA)
def test_shape(n_samples, family, data):
    shape = data.shape

    assert parametric_bootstrap(
        data, n_samples, family=family
    ).samples.shape == (
        n_samples,
        shape[0],
    )
    assert parametric_bootstrap(
        data, n_samples, family=family, statistic=safe_mean
    ).samples.shape == (n_samples,)


def test_empty_dataset():
    empty_dataset = np.arange(0)

    with pytest.raises(
        EmptyDataError, match="Supplied data is empty! Nothing to sample"
    ):
        parametric_bootstrap(empty_dataset, 1, family="norm")


def test_scalar_dataset():
    scalar_dataset = np.array(0)

    with pytest.raises(
        ValueError, match="Parameter 'data' cannot be a scalar"
    ):
        parametric_bootstrap(scalar_dataset, 1, family="norm")


def test_1d_data():
    with pytest.raises(
        ValueError, match="Parametric samplers expect 1-d data."
    ):
        parametric_bootstrap(RANDOM_10x5, 1, family="norm")


def test_invalid_family():
    with pytest.raises(ValueError):
        parametric_bootstrap(RANDOM_10x1, 1, family="snorf")



import numpy as np
import pytest

from brad.models import ConfidenceInterval, Resample

from .helpers import RANDOM_10x1


def test_standard_error():
    resample = Resample(2.0, np.arange(5))

    assert np.isclose(resample.standard_error(), np.sqrt(2.5))
    assert np.isclose(resample.standard_error(ddof=0), np.sqrt(2))


def test_standard_error_1d_check():
    resample_non_scalar_data = Resample(np.zeros(3), np.zeros(3))

    with pytest.raises(ValueError):
        resample_non_scalar_data.standard_error()

    resample_2d_samples = Resample(0.0, np.zeros((2, 2)))

    with pytest.raises(ValueError):
        resample_2d_samples.standard_error()


@pytest.mark.parametrize("confidence_level", [0.9, 0.95, 0.99])
@pytest.mark.parametrize("kind", ["pivotal", "normal", "percentile"])
def test_confidence_interval(confidence_level, kind):
    resample = Resample(0.0, RANDOM_10x1)

    ci = resample.confidence_interval(
        confidence_level=confidence_level, kind=kind
    )

    assert isinstance(ci, ConfidenceInterval)
    assert ci.kind == kind
    assert ci.confidence_level == confidence_level


def test_confidence_interval_1d_check():
    resample_non_scalar_data = Resample(np.zeros(3), np.zeros(3))

    with pytest.raises(ValueError):
        resample_non_scalar_data.confidence_interval()

    resample_2d_samples = Resample(0.0, np.zeros((2, 2)))

    with pytest.raises(ValueError):
        resample_2d_samples.confidence_interval()


def test_confidence_interval_kind():
    resample = Resample(0.0, RANDOM_10x1)

    with pytest.raises(ValueError):
        resample.confidence_interval(kind="snorf")



from dataclasses import dataclass
from typing import Union

import numpy as np
from scipy.stats import norm

Numeric = Union[float, np.ndarray]


@dataclass(frozen=True)
class ConfidenceInterval:
    """
    Confidence interval computed from bootstrap samples.

    Attributes:
        lower: The lower endpoint of the interval.
        upper: The upper endpoint of the interval.
        kind: The method used to construct the interval. Will be one of
            `"pivotal"`, `"normal"`, and `"percentile"`.
        confidence_level: The confidence level of the interval.
    """

    lower: float
    upper: float
    kind: str
    confidence_level: float


@dataclass(frozen=True)
class Resample:
    """
    Dataclass used when returning resampled data.

    Attributes:
        data: Either the original data, _or_ the original data mapped with the
            supplied statistic.
        samples: Samples produced by the chosen method, mapped with the
            supplied statistic.
    """

    data: Numeric
    samples: np.ndarray

    def _1d_check(self, method: str):
        if np.array(self.data).ndim != 0 or self.samples.ndim != 1:
            raise ValueError(
                f"{method} can only be calculated for a scalar statistic."
            )

    def standard_error(self, ddof: int = 1):
        """
        Compute the standard error of the samples.

        Args:
            ddof: "Delta degrees of freedom". The divisor used when calculating
                the standard deviation of the samples is `num_samples - ddof`.
                Defaults to `1` as per the original definition of the bootstrap
                standard error.
        """
        self._1d_check("Standard error")
        return self.samples.std(ddof=ddof)

    def confidence_interval(
        self, confidence_level=0.95, kind: str = "pivotal"
    ) -> ConfidenceInterval:
        """
        Compute a confidence interval from the samples. Can compute pivotal,
        percentile, or normal intervals.

        Args:
            confidence_level: The confidence level of the interval, stated as a
                number between `0` and `1`. If the confidence level is `0.95`
                (the default value) then the resulting confidence interval
                would contain the true value of the quantity being estimated
                95% of the time when calculated on different samples.
            kind: The type of interval to construct. Options are `"pivotal"`,
                `"percentile"` and `"normal"`.

        Returns:
            The upper and lower bounds for the interval.
        """
        self._1d_check("Confidence intervals")
        if kind.lower() == "pivotal":
            lower_quantile, upper_quantile = np.quantile(
                self.samples,
                [(1 - confidence_level) / 2, (1 + confidence_level) / 2],
            )
            return ConfidenceInterval(
                lower=2 * self.data - upper_quantile,
                upper=2 * self.data - lower_quantile,
                kind="pivotal",
                confidence_level=confidence_level,
            )
        elif kind.lower() == "normal":
            z = norm.ppf((1 + confidence_level) / 2)
            se = self.standard_error()
            return ConfidenceInterval(
                lower=self.data - z * se,
                upper=self.data + z * se,
                kind="normal",
                confidence_level=confidence_level,
            )
        elif kind.lower() == "percentile":
            quantiles = [
                (1 - confidence_level) / 2,
                (1 + confidence_level) / 2,
            ]
            lower, upper = np.quantile(self.samples, quantiles)
            return ConfidenceInterval(
                lower=lower,
                upper=upper,
                kind="percentile",
                confidence_level=confidence_level,
            )
        else:
            raise ValueError(
                f"Invalid option '{kind}' for argument kind. Should be one "
                "of: 'pivotal', 'normal', 'percentile'."
            )



import shutil
import warnings

import nox


def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return f"UserWarning: {msg}\n"


warnings.formatwarning = custom_formatwarning  # type: ignore

# don't run format session by default
nox.options.sessions = ["lint", "test"]

SOURCES = ["src", "tests", "noxfile.py"]


@nox.session()
def lint(session):
    """Lint all source code"""
    session.install("black", "flake8", "isort", "mypy")
    session.run("black", "--check", *SOURCES)
    session.run("flake8", *SOURCES)
    session.run("isort", "--check", *SOURCES)
    session.run("mypy", *SOURCES)


@nox.session(python=["3.7", "3.8", "3.9"])
def test(session):
    """Run tests"""
    session.install(".")
    session.install("pytest", "pytest-cov")
    session.run("pytest")


@nox.session(name="format")
def format_(session):
    """Format all source code"""
    session.install("black", "isort")
    session.run("black", *SOURCES)
    session.run("isort", *SOURCES)

    if shutil.which("prettier") is None:
        warnings.warn(
            "Prettier is not installed. Docs have not been formatted.\n\n"
            "Visit https://prettier.io/ to install.",
            stacklevel=0,
        )
    else:
        session.run("prettier", "--write", "docs", "README.md", external=True)


@nox.session()
def build(session):
    """Build package"""
    session.install(".", "build")
    session.run(
        "python", "-m", "build", "--sdist", "--wheel", "--outdir", "dist/"
    )


@nox.session(name="build-docs")
def build_docs(session):
    """Build documentation"""
    session.install(".")
    session.chdir("docs")
    session.install("-r", "requirements-docs.txt")
    session.run("mkdocs", "build")
